@@ -34,7 +34,7 @@ main(int argc, char *argv[])
 {
 	int ret;
 	char *table = "filter";
-	ip6tc_handle_t handle;
+	ip6tc_handle_t handle = NULL;
 
 	program_name = "ip6tables";
 	program_version = NETFILTER_VERSION;
